/**
 * Sort objects by object properties
 */
export function sortBy(arr, keys) {
    return arr.slice().sort((a, b) => {
        for (let i = 0; i < keys.length; i++) {
            const key = keys[i];
            if (a[key] < b[key])
                return -1;
            else if (a[key] > b[key])
                return 1;
        }
        return 0;
    });
}
/**
 * Take out the first N items of an array
 */
export function take(arr, n) {
    return arr.slice(0, n);
}
/**
 * Get the first element of an array
 */
export function head(arr) {
    return arr[0];
}
/**
 * Get multiple object or array value by paths. Example path: `a[0].b.c`
 */
export function at(source, paths) {
    const out = [];
    if (typeof paths === "string") {
        paths = [paths];
    }
    for (let i = 0; i < paths.length; i++) {
        const path = paths[i];
        const parts = path.split(/[.[\]]/g);
        let haystack = source;
        let found = true;
        for (let i = 0; i < parts.length; i++) {
            const part = parts[i];
            if (part === "")
                continue;
            if (!(part in haystack)) {
                found = false;
                break;
            }
            haystack = haystack[part];
        }
        if (found) {
            out.push(haystack);
        }
    }
    return out;
}
/**
 * Return item by path or default value if not present
 */
export function get(source, path, defaultValue) {
    if (Array.isArray(path)) {
        path = path.join(".");
    }
    const parts = path.split(/[.[\]]/g);
    let haystack = source;
    for (let i = 0; i < parts.length; i++) {
        const part = parts[i];
        if (part === "")
            continue;
        if (!(part in haystack))
            return defaultValue;
        haystack = haystack[part];
    }
    return haystack;
}
/**
 * Check if an item is in an object by path
 */
export function has(source, path) {
    const empty = "__tlodash_not_found";
    return get(source, path, empty) !== empty;
}
/**
 * Check if a collection (Array, Map, Set, Object) is empty
 */
export function isEmpty(v) {
    if (Array.isArray(v)) {
        return v.length === 0;
    }
    else if (v instanceof Set || v instanceof Map) {
        return v.size === 0;
    }
    else if (v !== null && typeof v === "object") {
        return Object.keys(v).length === 0;
    }
    return true;
}
/**
 * Pick values out of an object by predicate and return them in a new object
 */
export function pickBy(obj, predicate) {
    const out = {};
    Object.keys(obj).forEach(key => {
        if (predicate(obj[key], key)) {
            out[key] = obj[key];
        }
    });
    return out;
}
/**
 * Loop over a collection and call the callback on each item
 */
export function forEach(value, callback) {
    if (Array.isArray(value)) {
        value.forEach(callback);
    }
    else if (value instanceof Set || value instanceof Map) {
        value.forEach(callback);
    }
    else if (value !== null && typeof value === "object") {
        Object.keys(value).forEach(key => {
            callback(value[key], key, value);
        });
    }
}
/**
 * Convert string to kebab case
 */
export function kebabCase(input) {
    let out = input
        .replace(/[-._\s]+/g, "-")
        .replace(/[A-Z0-9]/, "-$&")
        .replace(/[-]{2,}/, "");
    if (out[0] === "-") {
        out = out.slice(1);
    }
    if (out[out.length - 1] === "-") {
        out = out.slice(0, -1);
    }
    return out.toLowerCase();
}
/**
 * Get index of an item in an array
 */
export function indexOf(arr, value, startIndex = 0) {
    if (startIndex === 0)
        return arr.indexOf(value);
    for (let i = startIndex; i < arr.length; i++) {
        if (arr[i] === value)
            return i;
    }
    return -1;
}
/**
 * Flatten an array by a single level
 */
export function flatten(arr) {
    return arr.reduce((acc, val) => acc.concat(val), []);
}
/**
 * Remove falsy values from array
 */
export function compact(arr) {
    return arr.filter(x => !!x);
}
/** Used to generate unique IDs. */
const idCounter = {};
/**
 * Generates a unique ID
 */
export function uniqueId(prefix = "$smoldash$") {
    const id = (idCounter[prefix] = (idCounter[prefix] || 0) + 1);
    return "" + (prefix === "$smoldash$" ? "" : prefix) + id;
}
// eslint-disable-next-line @typescript-eslint/ban-types
export function once(fn) {
    let result;
    let called = false;
    return function once_wrapped(...args) {
        if (!called) {
            called = true;
            result = fn.apply(this, args);
        }
        return result;
    };
}
/**
 * Create a duplicate free version of an array by a user iteratee
 */
export function uniqBy(arr, iteratee) {
    const seen = new Map();
    const type = typeof iteratee;
    arr.forEach(item => {
        let key = iteratee;
        if (type === "string" && item !== null && typeof item === "object") {
            key = item[iteratee];
        }
        else if (type === "function") {
            key = iteratee(item);
        }
        if (!seen.has(key)) {
            seen.set(key, item);
        }
    });
    return Array.from(seen.values());
}
/**
 * Pass the result of the first function to the next one.
 */
export function flow(...fns) {
    return function (...args) {
        let result = fns[0].apply(this, args);
        for (let i = 1; i < fns.length; i++) {
            result = fns[i].call(this, result);
        }
        return result;
    };
}
/**
 * Shallow clone of a value
 */
export function clone(value) {
    if (Array.isArray(value)) {
        return value.slice();
    }
    else if (value instanceof RegExp) {
        return new RegExp(value.source, value.flags);
    }
    else if (value instanceof Set) {
        return new Set(value);
    }
    else if (value instanceof Map) {
        return new Map(value);
    }
    else if (value instanceof Date) {
        return new Date(value);
    }
    else if (typeof value === "object" && value !== null) {
        return Object.assign({}, value);
    }
    return value;
}
/**
 * Deeply clone a value
 */
export function cloneDeep(value) {
    if (Array.isArray(value)) {
        return value.slice().map(cloneDeep);
    }
    else if (value instanceof RegExp) {
        return new RegExp(value.source, value.flags);
    }
    else if (value instanceof Set) {
        const out = new Set();
        value.forEach(v => out.add(cloneDeep(v)));
        return out;
    }
    else if (value instanceof Map) {
        const out = new Map();
        value.forEach((v, k) => out.set(k, cloneDeep(v)));
        return out;
    }
    else if (value instanceof Date) {
        return new Date(value);
    }
    else if (typeof value === "object" && value !== null) {
        const out = {};
        for (const k in value) {
            out[k] = cloneDeep(value[k]);
        }
        return out;
    }
    return value;
}
/**
 * Deeply merge multiple objects
 */
export function merge(...objs) {
    const a = objs[0];
    for (let i = 1; i < objs.length; i++) {
        const b = objs[i];
        for (const k in b) {
            if (a[k] !== null &&
                b[k] !== null &&
                typeof a[k] === "object" &&
                typeof b[k] === "object") {
                merge(a[k], b[k]);
            }
            else {
                a[k] = b[k];
            }
        }
    }
    return objs[0];
}
/**
 * Iterate the collection and return the element where the predicate returns true
 */
export function find(collection, predicate = x => !!x, fromIndex = 0) {
    let fn;
    if (typeof predicate === "string") {
        fn = (item) => !!item[predicate];
    }
    else if (Array.isArray(predicate)) {
        fn = (item) => item[predicate[0]] === predicate[1];
    }
    else if (typeof predicate === "object") {
        fn = (item) => {
            return Object.keys(predicate).every(v => !(v in predicate) || item[v] === predicate[v]);
        };
    }
    else if (typeof predicate === "function") {
        fn = predicate;
    }
    for (let i = fromIndex; i < collection.length; i++) {
        if (fn(collection[i])) {
            return collection[i];
        }
    }
}
/**
 * Shallowly compare to values
 */
export function isEqual(a, b) {
    if (a === null || b === null) {
        return a === b;
    }
    else if (Array.isArray(a) && Array.isArray(b)) {
        if (a.length !== b.length) {
            return false;
        }
        for (let i = 0; i < a.length; i++) {
            if (a[i] !== b[i]) {
                return false;
            }
        }
        return true;
    }
    else if (a instanceof RegExp && b instanceof RegExp) {
        return "" + a === "" + b;
    }
    else if (a instanceof Set && b instanceof Set) {
        if (a.size !== b.size)
            return false;
        for (const v of a.values()) {
            if (!b.has(v))
                return false;
        }
        return true;
    }
    else if (a instanceof Map && b instanceof Map) {
        if (a.size !== b.size)
            return false;
        for (const [k, v] of a.entries()) {
            if (!b.has(k))
                return false;
            if (b.get(k) !== v)
                return false;
        }
        return true;
    }
    else if (a instanceof Date && b instanceof Date) {
        return +a === +b;
    }
    else if (typeof a === "object" && typeof b === "object") {
        for (const i in a)
            if (!(i in b))
                return false;
        for (const i in b)
            if (a[i] !== b[i])
                return false;
        return true;
    }
    return a === b;
}
